//  UTop.cpp version 1.5
//  yudit package - Unicode Editor for the X Window System (and Linux) 
//
//  Author: gsinai@iname.com (Gaspar Sinai)
//  GNU Copyright (C) 1997,1998,1999  Gaspar Sinai
// 
//  yudit version 1.6  Copyright(C) 09 January,  2001, Tokyo Japan  Gaspar Sinai
//  yudit version 1.5  Copyright(C) 30 November, 1999, Tokyo Japan  Gaspar Sinai
//  yudit version 1.4  Copyright(C) 25 November, 1999, Tokyo Japan  Gaspar Sinai
//  yudit version 1.3  Copyright(C)  5 April,    1999, Tokyo Japan  Gaspar Sinai
//  yudit version 1.2  Copyright(C) 10 December, 1998, Tokyo Japan  Gaspar Sinai
//  yudit version 1.1  Copyright(C) 23 August,   1998, Tokyo Japan  Gaspar Sinai
//  yudit version 1.0  Copyright(C) 17 May,      1998, Tokyo Japan  Gaspar Sinai
//  yudit version 0.99 Copyright(C)  4 April,    1998, Tokyo Japan  Gaspar Sinai
//  yudit version 0.97 Copyright(C)  4 February, 1998, Tokyo Japan  Gaspar Sinai
//  yudit version 0.95 Copyright(C) 10 January,  1998, Tokyo Japan  Gaspar Sinai
//  yudit version 0.94 Copyright(C) 17 December, 1997, Tokyo Japan  Gaspar Sinai
//  yudit version 0.9 Copyright (C)  8 December, 1997, Tokyo Japan  Gaspar Sinai
//  yutex version 0.8 Copyright (C)  5 November, 1997, Tokyo Japan  Gaspar Sinai
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//
// The top object that is attached to the display and screen.
//

#include <iostream.h>
#include "UTop.h"
#include "UTextMenu.h"
#include <sys/types.h>
#include <sys/time.h>

UTop::UTop (void)
{
	init ();
}

UTop::~UTop()
{
	if (display!=0)
	{
		if (clipOwner!=None)
		{
			XSetSelectionOwner (display, XA_PRIMARY, None, 
				CurrentTime);
		}
		if (clipBuffer) delete[] clipBuffer;
		XCloseDisplay (display);
	}
}

void
UTop::init()
{
	display = XOpenDisplay (0);
	if (display == 0)
	{
		cerr << "error: XOpenDisplay\n";
		return;
	}
	screen = DefaultScreen (display);
	root = RootWindow (display, screen);
	border= 0;
	background = WhitePixel (display, screen);
	foreground = BlackPixel (display, screen);
	rectangle.x = 0;
	rectangle.y = 0;
	rectangle.width = 400;
	rectangle.height = 300;
	wmProtocols = XInternAtom (display, "WM_PROTOCOLS", False);
	wmDeleteWindow = XInternAtom (display, "WM_DELETE_WINDOW", False);
}

void
UTop::addComponent (UComponent *shell_)
{
	component.addItem ((unsigned long) shell_->window, (void*) shell_);
}

UComponent*
UTop::findComponent (Window w)
{
	return (UComponent*) (component.findItem ((unsigned long) w));
}

void
UTop::deleteComponent (UComponent *shell_)
{
	if (shell_->window==clipOwner) clipOwner =None;
	component.deleteItem ((unsigned long) shell_->window);
}

void
UTop::addShell (UComponent *shell_)
{
	shell.addItem ((unsigned long) shell_->window, (void*) shell_);
}

void
UTop::deleteShell (UComponent *shell_)
{
	shell.deleteItem ((unsigned long) shell_->window);
}

void
UTop::addEventHandler (UEventHandler* handler, unsigned long mask)
{
	eventHandler.addItem (mask, handler);
}

void
UTop::removeEventHandler (UEventHandler* handler, unsigned long mask)
{
	eventHandler.deleteItem (mask, handler);
}

const unsigned char*
UTop::getClipboard (Window requestor, int* len)
{

	Window		owner;
	unsigned char*	tmpBuffer;
	int		tmpLength;

	// If we own the selection ourselves, no need for interprocess
	// communication.
	if (clipOwner==(owner=XGetSelectionOwner (display, XA_PRIMARY)))
	{
		if (len != NULL) *len = clipLength;
		return clipBuffer;
	}
	if (clipBuffer) delete[] clipBuffer;
	clipBuffer = NULL;
	clipLength = 0;
	clipOwner = None;

	// Try UTF8_STRING first.
	tmpBuffer = getClipboard (XInternAtom(display,"UTF8_STRING",False),
				  requestor, &tmpLength);
	if (tmpBuffer != NULL)
	{
		clipBuffer = tmpBuffer;
		clipLength = tmpLength;
		if (len != NULL) *len = clipLength;
		return clipBuffer;
	}

	// Then try COMPOUND_TEXT.
#if X_HAVE_UTF8_STRING
	tmpBuffer = getClipboard (XInternAtom(display,"COMPOUND_TEXT",False),
				  requestor, &tmpLength);
	if (tmpBuffer != NULL)
	{
		XTextProperty textprop;
		char** list;
		int listcount;

		textprop.value = tmpBuffer;
		textprop.encoding = XInternAtom(display,"COMPOUND_TEXT",False);
		textprop.format = 8;
		textprop.nitems = tmpLength;
		if (Xutf8TextPropertyToTextList (display, &textprop,
						 &list, &listcount) == Success)
		{
			if (listcount == 1)
			{
				clipLength = strlen(list[0]);
				clipBuffer = new unsigned char[clipLength+1];
				memcpy (clipBuffer, list[0], clipLength);
				clipBuffer[clipLength] = '\0';
				XFreeStringList (list);
				delete[] tmpBuffer;
				if (len != NULL) *len = clipLength;
				return clipBuffer;
			}
			XFreeStringList (list);
		}
		delete[] tmpBuffer;
	}

	// Finally try STRING.
	tmpBuffer = getClipboard (XA_STRING, requestor, &tmpLength);
	if (tmpBuffer != NULL)
	{
		XTextProperty textprop;
		char** list;
		int listcount;

		textprop.value = tmpBuffer;
		textprop.encoding = XA_STRING;
		textprop.format = 8;
		textprop.nitems = tmpLength;
		if (Xutf8TextPropertyToTextList (display, &textprop,
						 &list, &listcount) == Success)
		{
			if (listcount == 1)
			{
				clipLength = strlen(list[0]);
				clipBuffer = new unsigned char[clipLength+1];
				memcpy (clipBuffer, list[0], clipLength);
				clipBuffer[clipLength] = '\0';
				XFreeStringList (list);
				delete[] tmpBuffer;
				if (len != NULL) *len = clipLength;
				return clipBuffer;
			}
			XFreeStringList (list);
		}
		delete[] tmpBuffer;
	}
#else
        /* FIXME: Yudit needs to set the encoding for this. 
           It may return non-utf8 strings */
	tmpBuffer = getClipboard (XInternAtom(display,"COMPOUND_TEXT",False),
				  requestor, &tmpLength);
	if (tmpBuffer != NULL)
	{
		clipBuffer = tmpBuffer;
		clipLength = tmpLength;
		if (len != NULL) *len = clipLength;
		return clipBuffer;
	}

	tmpBuffer = getClipboard (XA_STRING, requestor, &tmpLength);
	if (tmpBuffer != NULL)
	{
		clipBuffer = tmpBuffer;
		clipLength = tmpLength;
		if (len != NULL) *len = clipLength;
		return clipBuffer;
	}
#endif

	return NULL;
}

unsigned char*
UTop::getClipboard (Atom textType, Window requestor, int* len)
{
	XEvent		xevent;

	time_t		start;
	int		count;
	Atom		propty;
	int		propformat;
	unsigned long 	propsize;
	unsigned long 	rest;
	unsigned char*	propvalue;
	Atom		proptype;
	unsigned char*	buffer;
	unsigned char*	result;

	propty = XInternAtom (display, "YUDIT_SELECTION", False);

	XDeleteProperty (display, requestor, propty);
	XConvertSelection (display, XA_PRIMARY, textType, 
		propty, requestor, CurrentTime );

	// This is a bug... I dont know why but propty gets wrong
        // in the routine above on X11R5.
	propty = XInternAtom (display, "YUDIT_SELECTION", False);

	XFlush (display);
	for (start = time (NULL); ; )
	{
		if (XCheckTypedWindowEvent(display, requestor,
			SelectionNotify, &xevent))
			break;
		if (time (NULL) - start >= 8)
		{
			cerr << "warn: clipboard timeout.\n";
			return NULL;
		}
	}

	// FIXME: This does not deal with INCR properties.

	result = NULL;
	count = 0;
	// Read the propty from the selection window
	while (Success==XGetWindowProperty (display, 
			xevent.xselection.requestor, 
			propty,
			count, 100000L, True, AnyPropertyType,
			&proptype,  &propformat, &propsize, &rest, &propvalue))
	{
		if (propvalue==NULL) break;
		if ((proptype != textType) || propsize==0 || propformat !=8)
		{
			XFree (propvalue);
			break;
		}
		buffer = new unsigned char[propsize+count+4];
		if (count)
		{
			memcpy (buffer, result, count);
			delete[] result;
		}
		result = buffer;
		memcpy (&result[count], propvalue, propsize);
		XFree (propvalue);
		count += propsize;
		if (rest==0) break;
	}
	if (result)
	{
		result[count] = 0;
	}
	if (len !=0) *len = count;
	return result;
}

void
UTop::putClipboard (Window win, const unsigned char*in_, int len)
{
	UComponent*	comp;
	if (clipOwner != win && clipOwner != None)
	{
		comp = findComponent (clipOwner);
		if (comp) comp->clearSelection();
	}
	XSetSelectionOwner (display, XA_PRIMARY, win, CurrentTime);
	if (XGetSelectionOwner (display, XA_PRIMARY) != win)
	{
		// Oops
		clipOwner = None;
		return;
	}
	clipOwner = win;
	if (clipBuffer) delete[] clipBuffer;
	clipBuffer = new unsigned char [len+4];
	memcpy (clipBuffer, in_, len);
	clipLength = len;
	clipBuffer[len] = 0;
}

void
UTop::clipEvent (XEvent* event)
{
	UComponent* 			comp;
	XEvent				xevent;
	XSelectionRequestEvent&		sel = event->xselectionrequest;

	switch (event->type)
	{
	case SelectionClear:
		comp = findComponent (clipOwner);
		if (comp) comp->clearSelection();
		clipOwner = None;
		return;
	case SelectionNotify:
		clipOwner = None;
		return;
	case SelectionRequest:
		/* FIXME: This does not answer MULTIPLE and TIMESTAMP requests. */
		if (sel.target == XInternAtom(display,"TARGETS",False))
		{
			Atom* targets = new Atom[5];
			targets[0] = XInternAtom(display,"TARGETS",False);
			targets[1] = XInternAtom(display,"TEXT",False);
			targets[2] = XA_STRING;
			targets[3] = XInternAtom(display,"COMPOUND_TEXT",False);
			targets[4] = XInternAtom(display,"UTF8_STRING",False);
			XChangeProperty (sel.display,
					 sel.requestor, sel.property,
					 XA_ATOM, 32, PropModeReplace,
					 (unsigned char *) targets, 5);
			delete[] targets;
		}
		else
		// Asked for TEXT or UTF8_STRING -> answer with UTF8_STRING.
		if (sel.target == XInternAtom(display,"TEXT",False)
		    || sel.target == XInternAtom(display,"UTF8_STRING",False))
		{
			XChangeProperty (sel.display,
					 sel.requestor, sel.property,
					 XInternAtom(display,"UTF8_STRING",False), 8, PropModeReplace,
					 clipBuffer, clipLength);
		}
		else
#if X_HAVE_UTF8_STRING
		// Asked for COMPOUND_TEXT -> answer with COMPOUND_TEXT if convertible.
		if (sel.target == XInternAtom(display,"COMPOUND_TEXT",False))
		{
			XTextProperty textprop;

			if (Xutf8TextListToTextProperty(display, (char **) &clipBuffer, 1,
							XCompoundTextStyle,
							&textprop) == Success)
			{
				XChangeProperty (sel.display,
						 sel.requestor, sel.property,
						 sel.target, 8, PropModeReplace,
						 textprop.value, textprop.nitems);
			} else {
				sel.property = None;
			}
		}
		else
		// Asked for STRING -> answer with STRING if convertible.
		if (sel.target == XA_STRING)
		{
			XTextProperty textprop;

			if (Xutf8TextListToTextProperty(display, (char **) &clipBuffer, 1,
							XStringStyle,
							&textprop) == Success)
			{
				XChangeProperty (sel.display,
						 sel.requestor, sel.property,
						 XA_STRING, 8, PropModeReplace,
						 textprop.value, textprop.nitems);
			} else {
				sel.property = None;
			}
		}
		else
#else
               /* FIXME Yudit needs to set the encoding for this. */
		if (sel.target == XInternAtom(display,"COMPOUND_TEXT",False)
                  || sel.target == XA_STRING)
                {
			XChangeProperty (sel.display,
				 sel.requestor, sel.property,
                                 sel.target,
				8, PropModeReplace,
				 clipBuffer, clipLength);
                }
                else

#endif
		{
			sel.property = None;
		}

		xevent.xselection.type = SelectionNotify;
		xevent.xselection.display = sel.display;
		xevent.xselection.requestor = sel.requestor;
		xevent.xselection.selection = sel.selection;
		xevent.xselection.target = sel.target;
		xevent.xselection.property = sel.property;
		xevent.xselection.time = sel.time;
		XSendEvent (sel.display, sel.requestor, False, 0, &xevent);
		return;
	default:
		break;
	}
}
